(in-package :araneida)

(defvar *handler-timeout* 60) ;seconds

(defvar *sockets* nil)
(defvar *threads* nil)
(defvar *serve-event-handlers* nil)

(defun accept-handler (listener)
  (let ((socket (socket-accept listener)))
    (handle-a-request
     (socket-make-stream socket :element-type 'base-char
			 :input t :output t :buffering :full)
     socket)))

(defun accept-with-timeout-handler (listener)
  (handler-case
      (let ((socket (socket-accept listener)))
        (sb-ext:with-timeout *handler-timeout*
          (handle-a-request
           (socket-make-stream socket :element-type 'base-char
                               :input t :output t :buffering :full)
           socket)))
    (sb-bsd-sockets::bad-file-descriptor-error (c)
      (format t "accept: bad file descriptor error (socket shut down?)~%" c)
      (return-from accept-with-timeout-handler nil))
    (sb-bsd-sockets::socket-error (c)
      (format t "socket error ~A~%" c))
    (sb-ext:timeout (c)
      (format t "timed out processing request: ~A~%" c))))

(defun install-handlers-aux ()
  (let* ((ports (remove-duplicates
                 (mapcar #'server-port *exported-servers*)))
         (threads nil)
         (listeners
          (mapcar
           (lambda (port)
             (let ((socket
                    (make-instance 'inet-socket :type :stream :protocol :tcp)))
               (setf (sockopt-reuse-address socket) t)
               (socket-bind socket #(0 0 0 0) port)
               (socket-listen socket 15)
               socket))
           ports)))
    (or listeners (error "No sockets to listen on"))))

(defun install-thread-handlers (&optional (num-threads 5))
  (setf *sockets* (install-handlers-aux))
  (dolist (socket *sockets*)
    (dotimes (n num-threads)
      (push (sb-thread:make-thread
	     (lambda ()
	       (loop
		(with-simple-restart
		    (abort-response
		     "Abort this response and answer another request")
		  (accept-with-timeout-handler socket)))))
	    *threads*))))

(defun install-serve-event-handlers ()
  (setf *sockets* (install-handlers-aux))
  (dolist (socket *sockets*)
    (push 
     (sb-sys:add-fd-handler (sb-bsd-sockets:socket-file-descriptor  socket)
			    :input (lambda (fd) (accept-handler socket)))
     *serve-event-handlers*)))

(defun remove-handlers ()
  (dolist (th *threads*)
    (sb-thread:destroy-thread th))
  (dolist (h *serve-event-handlers*)
    (sb-sys:remove-fd-handler h))
  (dolist (s *sockets*)
    (sb-bsd-sockets:socket-close s)))


(eval-when (:load-toplevel :execute)
  (sb-sys:ignore-interrupt :sigpipe))

(defun handle-a-request (stream socket)
  (let* ((request-line (read-folded-line stream nil))
         (headers (read-headers stream))
         (content-length (parse-integer
                          (or (header-value :content-length headers) "0")))
	 (body (and (> content-length 0)
		    (make-array content-length :element-type
				(stream-element-type stream))))
         (len (and (> content-length 0)
		   (read-sequence body stream))))
    (destructuring-bind
	  (method url-string protocol) (split request-line 3 '(#\Space))
      (declare (ignore protocol))
      (let* ((parsed-body (if body (parse-body body '(#\&) len) nil))
             (url (merge-url
                   (server-base-url (server-for-name-colon-port
                                     (header-value :host headers)))
                   url-string))
             (r (make-instance
                 'request :url url
		 :urlstring (urlstring url)
                 :method (intern (nstring-upcase method) 'keyword)
                 :body parsed-body
                 :stream stream :headers headers :socket socket)))
	(handle-request *root-handler* r)
	;; when we support persistent connections, we'll have to remove
	;; this close operation
	(let* ((s (request-stream r)))
	  (forcibly-close-stream s))))))

(defun parse-body (body-string &optional (delimiters '(#\&)) end)
  "Parse BODY-STRING returning list of (var val) pairs"
  (mapcar (lambda (x)
            (mapcar #'urlstring-unescape (split-sequence #\= x :count 2)))
          (split-sequence-if (lambda (x) (member x delimiters))
			     body-string :end end)))

(defun read-folded-line (stream &optional eof-error-p eof-value)
  "Read a complete logical header line, including folded continuation lines."
  (with-output-to-string (o)
    (loop
     (let* ((l (read-line stream eof-error-p eof-value))
	    (end (position-if (lambda (x) (or (eql x (code-char 10))
					      (eql x (code-char 13))))
			      l))
	    (next (and (if end (> end 0) t)
		       (> (length l) 0)
		       (peek-char nil stream nil nil))))
       (write-sequence l o :end end)
       (unless (or (eql next #\Space) (eql next #\Tab)) (return))))))

;;; Unless you're me (and probably not even then), you should be
;;; using REQUEST-HEADER not this:

(defun header-value (name header-list)
  "Get the value of the header named NAME in HEADER-LIST"
  (cadr (assoc name header-list)))

(defun read-headers (stream)
  (let ((headers nil))
    (do ((line (read-folded-line stream t) (read-folded-line stream t)))
        ((or (not line) (zerop (length line))) headers)
      
      ;; RFC 1945: "Each header field consists of a name followed
      ;; immediately by a colon (":"), a single space (SP) character,
      ;; and the field value. Field names are case- insensitive."
      (let* ((colon-pos (position #\: line))
	     (keyword (subseq line 0 colon-pos))
	     (value (subseq line (+ 2 colon-pos)))
	     (keyword-symbol (intern (string-upcase keyword) :keyword))
	     (keyword-value (assoc keyword-symbol  headers)))
	(if keyword-value
	    (setf (cdr (last keyword-value)) (list value))
            (setf headers (acons keyword-symbol (list value) headers)))))
    headers))
        
