(in-package :araneida-examples)

;;; Araneida server contains the source, docs, and examples for Araneida

;; default HTTP port is correct if serving through Apache proxy; the
;; internal port can be used here if you want to talk to araneida
;; directly
(defvar *araneida-url* (make-instance 'http-url :host *araneida-site-name*
				      #| :port *araneida-internal-port* |#))
(defvar *araneida-server*
  (make-instance 'server
                 :name *araneida-site-name*
                 :base-url *araneida-url* :port *araneida-internal-port*))
(export-server *araneida-server*)


;;; example 1: a simple hit counter

(defvar *home-page-hits* 0)

(defun another-hit ()
  (princ-to-string (incf *home-page-hits*)))

;; this is the function that will be invoked when a request is made

(defun counter-handler (request rest-of-url)
  (request-send-headers request)
  (write-sequence
   (html
    `(html
      (head ((link :rel "stylesheet" :href "/docs/araneida/araneida.css"))
            (title "Araneida Web Counter"))
      (body
       (h1 "Araneida Web Counter")
       (p ,(urlstring (request-url request)))
       (p ,(another-hit) " hits to this page since Lisp was started"))))
   (request-stream request)))

;;; now we register our handler function with the dispatcher

(export-handler (merge-url *araneida-url* "/counter") 
                'counter-handler :match :prefix :method t :stage :response)

;;; example 2: an error

;;; the handler that causes an error
(export-handler (merge-url *araneida-url* "/error")
                (lambda (r rest) (/ 1 0)))

;;; the handler that catches it (and all other errors under
;;; *araneida-url*).  if *break-on-handler-errors* is set, we can
;;; interactively debug this: obviously that's not practical on a
;;; production site :-)

(defvar *break-on-handler-errors* nil)
(export-handler *araneida-url*
                (lambda (r rest)
                  (format t "Error: ~A~%" (request-condition r))
                  (request-send-error
                   r 500
                   (format nil "caught error: ~A~%~A"
                           (request-condition r)
                           (with-output-to-string (out)
                             #+cmu (debug:backtrace 40 out)
			     #+sbcl (sb-debug:backtrace 40 out))))
		  ;; XXX if this works, I don't see how.  there is no
		  ;; such restart
                  (unless *break-on-handler-errors*
                    (invoke-restart 'araneida::abort-request)))
                :stage :error)
                


;;; example 3: export the araneida source code as plain files

(export-handler *araneida-url*
                `(file-request-handler ,*araneida-source-directory*)
                :match :prefix :method t :stage :response)

;;; example 4: to demonstrate the authentication stuff, we export the
;;; source directory again under /private/, and install handlers for
;;; the :authentication and :authorization stages too

(export-handler (merge-url *araneida-url* "/private/")
                `(file-request-handler ,*araneida-source-directory*)
                :match :prefix :method t :stage :response)

(export-handler (merge-url *araneida-url* "/private/")
                (araneida::make-basic-authentication-handler
                 "Private documents"
                 (lambda (u p)
                   ;; XXX really ought to check that the password is
                   ;; in some way correct.
                   (format t "u ~a p ~a~%" u p)
                   ;; store the username-as-string in request-user
                   u))
                :match :prefix :method t :stage :authentication)

;;; example 5a,b: export the documentation at "/docs/", and redirect from
;;; the root url of the site to the documentation index

(export-handler (merge-url *araneida-url* "/docs/")
                `(file-request-handler
		  ,(merge-pathnames
		    (make-pathname :directory '(:relative :back "doc")
				   :name nil)
		    *load-truename*))
                :match :prefix :method t :stage :response)
(export-handler *araneida-url*
                (lambda (r rest)
                  (request-redirect
                   r (merge-url *araneida-url* "/docs/index.html")))
                :match :exact :method t :stage :response)

;;; example 6: how to use the pattern rewriting stuff.  You can also call
;;; rewrite-tree yourself: macroexpand the define-page form for an example

(define-patterns example-patterns
  (((page :any)
    :rewrite (html (head (title "Araneida: " ?title)) (body ?body)))
   ((h1 :any)
    :rewrite (p ((font :size "+2" :color "green") ?*)))
   ((red :any)
    :rewrite ((font :color red) ?*))
   ((p :any) :preserve)))

(define-page
 ((merge-url *araneida-url* "/rewritten")
  :patterns example-patterns
  :changed-time (get-universal-time))
 '(page
   (title "foo")
   (body (h1 "title") (p "hello " (red "red") " world")))) 

#|
;;; The static site contains any HTML you like; it's here mostly for
;;; demonstration purposes

(defvar *static-url* (make-instance 'http-url :host *static-site-name*))
(defvar *static-server*
  (make-instance 'server
                 :name *static-site-name*
		 :base-url *static-url* :port *araneida-internal-port*))
(export-server *static-server*)

(export-handler *static-url*
                `(file-request-handler ,*static-pages-directory*)
                :method t)

(export-handler *static-url*
                (lambda (r rest)
                  (format t "Error: ~A~%" (request-condition r))
                  (request-send-error r 500
                                      (format nil "~A" (request-condition r))))
                :stage :error)
                
|#

;;; OK, let's roll

(defun init ()
  "Write the apache config file segment, read the mime.types file.
Call this whenever restarting serving after changing the list of
exported servers"
  (setf araneida:*content-types*
        (araneida:read-mime-types *mime-types-file*))
  (with-open-file (conf *apache-config-segment*
                        :direction :output
                        :if-does-not-exist :create)
    (output-apache-conf conf))
  (format t "Check ~S to see if it makes sense, then start
or restart apache if it looks like it might have changed"
          *apache-config-segment*))

(defun start ()
  "Install the Araneida event handlers (start answering requests).
Call this whenever you like"
  (install-serve-event-handlers))

(defun stop ()
  "Remove the Araneida event handlers (stop answering requests).  Call
this whenever you like"
  (remove-serve-event-handlers))



